@@ -175,14 +175,43 @@ NotificationDaemon.prototype = {
         }
     },
 
-    _newSource: function(title, pid) {
-        let source = new Source(title, pid);
-        this._sources[pid] = source;
+    // Returns the source associated with ndata.notification if it is set.
+    // Otherwise, returns the source associated with the pid if one is
+    // stored in this._sources and the notification is not transient.
+    // Otherwise, creates a new source as long as pid is provided.
+    //
+    // Either a pid or ndata.notification is needed to retrieve or
+    // create a source.
+    _getSource: function(title, pid, ndata) {
+        if (!pid && !(ndata && ndata.notification))
+            return null;
 
-        source.connect('destroy', Lang.bind(this,
-            function() {
-                delete this._sources[pid];
-            }));
+        // We use notification's source for the notifications we still have
+        // around that are getting replaced because we don't keep sources
+        // for transient notifications in this._sources, but we still want
+        // the notification associated with them to get replaced correctly.
+        if (ndata && ndata.notification)
+            return ndata.notification.source;
+
+        let isForTransientNotification = (ndata && ndata.hints['transient'] == true);
+
+        // We don't want to override a persistent notification
+        // with a transient one from the same sender, so we
+        // always create a new source object for new transient notifications
+        // and never add it to this._sources .
+        if (!isForTransientNotification && this._sources[pid])
+            return this._sources[pid];
+
+        let source = new Source(title, pid);
+        source.setTransient(isForTransientNotification);
+
+        if (!isForTransientNotification) {
+            this._sources[pid] = source;
+            source.connect('destroy', Lang.bind(this,
+                function() {
+                    delete this._sources[pid];
+                }));
+        }
 
         Main.messageTray.add(source);
         return source;
@@ -234,7 +263,8 @@ NotificationDaemon.prototype = {
 
         let sender = DBus.getCurrentMessageContext().sender;
         let pid = this._senderToPid[sender];
-        let source = pid ? this._sources[pid] : null;
+
+        let source = this._getSource(appName, pid, ndata);
 
         if (source) {
             this._notifyForSource(source, ndata);
@@ -255,16 +285,23 @@ NotificationDaemon.prototype = {
                 if (!ndata)
                     return;
 
-                this._senderToPid[sender] = pid;
-                source = this._sources[pid];
-
-                if (!source)
-                    source = this._newSource(appName, pid);
-                source.connect('destroy', Lang.bind(this,
-                    function() {
-                        delete this._senderToPid[sender];
-                    }));
+                source = this._getSource(appName, pid, ndata);
 
+                // We only store sender-pid entries for persistent sources.
+                // Removing the entries once the source is destroyed
+                // would result in the entries associated with transient
+                // sources removed once the notification is shown anyway.
+                // However, keeping these pairs would mean that we would
+                // possibly remove an entry associated with a persistent
+                // source when a transient source for the same sender is
+                // distroyed.
+                if (!source.isTransient) {
+                    this._senderToPid[sender] = pid;
+                    source.connect('destroy', Lang.bind(this,
+                        function() {
+                            delete this._senderToPid[sender];
+                        }));
+                }
                 this._notifyForSource(source, ndata);
             }));
 
@@ -291,7 +328,10 @@ NotificationDaemon.prototype = {
                 function(n) {
                     delete this._notifications[id];
                 }));
-            notification.connect('action-invoked', Lang.bind(this, this._actionInvoked, source, id));
+            notification.connect('action-invoked', Lang.bind(this,
+                function(n, actionId) {
+                    this._emitActionInvoked(id, actionId);
+                }));
         } else {
             notification.update(summary, body, { icon: iconActor,
                                                  bannerMarkup: true,
@@ -303,8 +343,21 @@ NotificationDaemon.prototype = {
             for (let i = 0; i < actions.length - 1; i += 2)
                 notification.addButton(actions[i], actions[i + 1]);
         }
-
-        notification.setUrgent(hints.urgency == Urgency.CRITICAL);
+        switch (hints.urgency) {
+            case Urgency.LOW:
+                notification.setUrgency(MessageTray.Urgency.LOW);
+                break;
+            case Urgency.NORMAL:
+                notification.setUrgency(MessageTray.Urgency.NORMAL);
+                break;
+            case Urgency.CRITICAL:
+                notification.setUrgency(MessageTray.Urgency.CRITICAL);
+                break;
+        }
+        notification.setResident(hints.resident == true);
+        // 'transient' is a reserved keyword in JS, so we have to retrieve the value
+        // of the 'transient' hint with hints['transient'] rather than hints.transient
+        notification.setTransient(hints['transient'] == true);
 
         let sourceIconActor = source.useNotificationIcon ? this._iconForNotificationData(icon, hints, source.ICON_SIZE) : null;
         source.notify(notification, sourceIconActor);
@@ -352,17 +405,13 @@ NotificationDaemon.prototype = {
         for (let id in this._sources) {
             let source = this._sources[id];
             if (source.app == tracker.focus_app) {
-                source.activated();
+                if (source.notification && !source.notification.resident)
+                    source.notification.destroy();
                 return;
             }
         }
     },
 
-    _actionInvoked: function(notification, action, source, id) {
-        source.activated();
-        this._emitActionInvoked(id, action);
-    },
-
     _emitNotificationClosed: function(id, reason) {
         DBus.session.emit_signal('/org/freedesktop/Notifications',
                                  'org.freedesktop.Notifications',
@@ -378,9 +427,7 @@ NotificationDaemon.prototype = {
     },
 
     _onTrayIconAdded: function(o, icon) {
-        let source = this._sources[icon.pid];
-        if (!source)
-            source = this._newSource(icon.title || icon.wm_class || _("Unknown"), icon.pid);
+        let source = this._getSource(icon.title || icon.wm_class || _("Unknown"), icon.pid, null);
         source.setTrayIcon(icon);
     },
 
@@ -421,13 +468,16 @@ Source.prototype = {
     },
 
     _setApp: function() {
+        if (this.app)
+            return;
+
         this.app = Shell.WindowTracker.get_default().get_app_from_pid(this._pid);
         if (!this.app)
             return;
 
         // Only override the icon if we were previously using
-        // notification-based icons (ie, not a trayicon)
-        if (this.useNotificationIcon) {
+        // notification-based icons (ie, not a trayicon) or if it was unset before
+        if (!this._isTrayIcon) {
             this.useNotificationIcon = false;
             this._setSummaryIcon(this.app.create_icon_texture (this.ICON_SIZE));
         }
@@ -440,12 +490,10 @@ Source.prototype = {
     },
 
     _notificationClicked: function(notification) {
-        notification.destroy();
         this.openApp();
-        this.activated();
     },
 
-    activated: function() {
+    _notificationRemoved: function() {
         if (!this._isTrayIcon)
             this.destroy();
     },
