@@ -26,6 +26,8 @@ const PANEL_ICON_SIZE = 24;
 
 const HOT_CORNER_ACTIVATION_TIMEOUT = 0.5;
 
+const BUTTON_DND_ACTIVATION_TIMEOUT = 250;
+
 const ANIMATED_ICON_UPDATE_TIMEOUT = 100;
 const SPINNER_UPDATE_TIMEOUT = 130;
 const SPINNER_SPEED = 0.02;
@@ -40,8 +42,10 @@ const STANDARD_TRAY_ICON_SHELL_IMPLEMENTATION = {
 if (Config.HAVE_BLUETOOTH)
     STANDARD_TRAY_ICON_SHELL_IMPLEMENTATION['bluetooth'] = imports.ui.status.bluetooth.Indicator;
 
-const CLOCK_FORMAT_KEY        = 'format';
-const CLOCK_CUSTOM_FORMAT_KEY = 'custom-format';
+// in org.gnome.desktop.interface
+const CLOCK_FORMAT_KEY        = 'clock-format';
+
+// in org.gnome.shell.clock
 const CLOCK_SHOW_DATE_KEY     = 'show-date';
 const CLOCK_SHOW_SECONDS_KEY  = 'show-seconds';
 
@@ -492,36 +496,32 @@ function ClockButton() {
 }
 
 ClockButton.prototype = {
-    __proto__: PanelMenu.Button.prototype,
-
     _init: function() {
-        PanelMenu.Button.prototype._init.call(this, St.Align.START);
-        this.menu.addAction(_("Preferences"), Lang.bind(this, this._onPrefs));
+        this.actor = new St.Bin({ style_class: 'panel-button',
+                                  reactive: true,
+                                  can_focus: true,
+                                  x_fill: true,
+                                  y_fill: false,
+                                  track_hover: true });
+        this.actor._delegate = this;
+        this.actor.connect('button-press-event',
+                           Lang.bind(this, this._toggleCalendar));
 
         this._clock = new St.Label();
         this.actor.set_child(this._clock);
 
         this._calendarPopup = null;
 
+        this._desktopSettings = new Gio.Settings({ schema: 'org.gnome.desktop.interface' });
         this._clockSettings = new Gio.Settings({ schema: 'org.gnome.shell.clock' });
+
+        this._desktopSettings.connect('changed', Lang.bind(this, this._updateClock));
         this._clockSettings.connect('changed', Lang.bind(this, this._updateClock));
 
         // Start the clock
         this._updateClock();
     },
 
-    _onButtonPress: function(actor, event) {
-        let button = event.get_button();
-        let menuShowing = this.menu.isOpen;
-        let calendarShowing = this._calendarPopup && this._calendarPopup.isOpen;
-
-        if (menuShowing || (button == 3 && !calendarShowing))
-            this.menu.toggle();
-        else
-            this._toggleCalendar();
-        return true;
-    },
-
     closeCalendar: function() {
         if (!this._calendarPopup || !this._calendarPopup.isOpen)
             return;
@@ -537,24 +537,12 @@ ClockButton.prototype = {
         this.actor.add_style_pseudo_class('pressed');
     },
 
-    _onPrefs: function() {
-        let args = ['gnome-shell-clock-preferences'];
-        let p = new Shell.Process({ args: args });
-
-        p.run();
-    },
-
     _toggleCalendar: function() {
         if (this._calendarPopup == null) {
             this._calendarPopup = new CalendarPopup();
             this._calendarPopup.actor.hide();
         }
 
-        if (this.menu.isOpen && !this._calendarPopup.isOpen) {
-            this.menu.close();
-            return;
-        }
-
         if (!this._calendarPopup.isOpen)
             this.openCalendar();
         else
@@ -562,23 +550,13 @@ ClockButton.prototype = {
     },
 
     _updateClock: function() {
-        let format = this._clockSettings.get_string(CLOCK_FORMAT_KEY);
+        let format = this._desktopSettings.get_string(CLOCK_FORMAT_KEY);
         let showDate = this._clockSettings.get_boolean(CLOCK_SHOW_DATE_KEY);
         let showSeconds = this._clockSettings.get_boolean(CLOCK_SHOW_SECONDS_KEY);
 
         let clockFormat;
         switch (format) {
-            case 'unix':
-                // force updates every second
-                showSeconds = true;
-                clockFormat = '%s';
-                break;
-            case 'custom':
-                // force updates every second
-                showSeconds = true;
-                clockFormat = this._clockSettings.get_string(CLOCK_CUSTOM_FORMAT_KEY);
-                break;
-            case '24-hour':
+            case '24h':
                 if (showDate)
 	            /* Translators: This is the time format with date used
                        in 24-hour mode. */
@@ -590,7 +568,7 @@ ClockButton.prototype = {
                     clockFormat = showSeconds ? _("%a %R:%S")
                                               : _("%a %R");
                 break;
-            case '12-hour':
+            case '12h':
             default:
                 if (showDate)
 	            /* Translators: This is a time format with date used
@@ -752,7 +730,20 @@ Panel.prototype = {
                                          reactive: true,
                                          can_focus: true });
         this.button.set_child(label);
-
+        this.button._delegate = this.button;
+        this.button._xdndTimeOut = 0;
+        this.button.handleDragOver = Lang.bind(this,
+            function(source, actor, x, y, time) {
+                 if (source == Main.xdndHandler) {
+                    if (this.button._xdndTimeOut != 0)
+                        Mainloop.source_remove(this.button._xdndTimeOut);
+                    this.button._xdndTimeOut = Mainloop.timeout_add(BUTTON_DND_ACTIVATION_TIMEOUT,
+                                                                    Lang.bind(this,
+                                                                                function() {
+                                                                                    this._xdndShowOverview(actor);
+                                                                                }));
+                 }
+            });
         this._leftBox.add(this.button);
 
         // We use this flag to mark the case where the user has entered the
@@ -790,6 +781,18 @@ Panel.prototype = {
         this._hotCorner.connect('leave-event',
                                 Lang.bind(this, this._onHotCornerLeft));
 
+        this._hotCorner._delegate = this._hotCorner;
+        this._hotCorner.handleDragOver = Lang.bind(this,
+            function(source, actor, x, y, time) {
+                 if (source == Main.xdndHandler) {
+                    if(!Main.overview.visible && !Main.overview.animationInProgress) {
+                        this.rippleAnimation();
+                        Main.overview.showTemporarily();
+                        Main.overview.beginItemDrag(actor);
+                    }
+                 }
+            });
+
         this._boxContainer.add_actor(this._hotCornerEnvirons);
         this._boxContainer.add_actor(this._hotCorner);
 
@@ -803,9 +806,6 @@ Panel.prototype = {
         this._clockButton = new ClockButton();
         this._centerBox.add(this._clockButton.actor, { y_fill: true });
 
-        let clockMenuManager = new PopupMenu.PopupMenuManager(this);
-        clockMenuManager.addMenu(this._clockButton.menu);
-
         /* right */
 
         // System status applets live in statusBox, while legacy tray icons
@@ -848,6 +848,25 @@ Panel.prototype = {
         Main.chrome.addActor(this.actor, { visibleInOverview: true });
     },
 
+    _xdndShowOverview: function (actor) {
+        let [x, y, mask] = global.get_pointer();
+        let pickedActor = global.stage.get_actor_at_pos(Clutter.PickMode.REACTIVE, x, y);
+
+        if (pickedActor != this.button) {
+            Mainloop.source_remove(this.button._xdndTimeOut);
+            this.button._xdndTimeOut = 0;
+            return;
+        }
+
+        if(!Main.overview.visible && !Main.overview.animationInProgress) {
+            Main.overview.showTemporarily();
+            Main.overview.beginItemDrag(actor);
+        }
+
+        Mainloop.source_remove(this.button._xdndTimeOut);
+        this.button._xdndTimeOut = 0;
+    },
+
     startStatusArea: function() {
         for (let i = 0; i < STANDARD_TRAY_ICON_ORDER.length; i++) {
             let role = STANDARD_TRAY_ICON_ORDER[i];
@@ -942,6 +961,17 @@ Panel.prototype = {
         Main.uiGroup.add_actor(ripple);
     },
 
+    rippleAnimation: function() {
+        // Show three concentric ripples expanding outwards; the exact
+        // parameters were found by trial and error, so don't look
+        // for them to make perfect sense mathematically
+
+        //              delay  time  scale opacity => scale opacity
+        this._addRipple(0.0,   0.83,  0.25,  1.0,    1.5,  0.0);
+        this._addRipple(0.05,  1.0,   0.0,   0.7,    1.25, 0.0);
+        this._addRipple(0.35,  1.0,   0.0,   0.3,    1,    0.0);
+    },
+
     _onHotCornerEntered : function() {
         if (this._menus.grabbed)
             return false;
@@ -950,14 +980,7 @@ Panel.prototype = {
             if (!Main.overview.animationInProgress) {
                 this._hotCornerActivationTime = Date.now() / 1000;
 
-                // Show three concentric ripples expanding outwards; the exact
-                // parameters were found by trial and error, so don't look
-                // for them to make perfect sense mathematically
-
-                //              delay  time  scale opacity => scale opacity
-                this._addRipple(0.0,   0.83,  0.25,  1.0,    1.5,  0.0);
-                this._addRipple(0.05,  1.0,   0.0,   0.7,    1.25, 0.0);
-                this._addRipple(0.35,  1.0,   0.0,   0.3,    1,    0.0);
+                this.rippleAnimation();
                 Main.overview.toggle();
             }
         }
@@ -1004,6 +1027,9 @@ function CalendarPopup() {
 CalendarPopup.prototype = {
     _init: function() {
         let panelActor = Main.panel.actor;
+        let alignConstraint = new Clutter.AlignConstraint({ source: panelActor,
+                                                            align_axis: Clutter.AlignAxis.X_AXIS,
+                                                            factor: 0.5 });
 
         this.actor = new St.Bin({ name: 'calendarPopup' });
 
@@ -1015,7 +1041,7 @@ CalendarPopup.prototype = {
         Main.chrome.addActor(this.actor, { visibleInOverview: true,
                                            affectsStruts: false });
         this.actor.y = (panelActor.y + panelActor.height - this.actor.height);
-        this.calendar.actor.connect('notify::width', Lang.bind(this, this._centerPopup));
+        this.actor.add_constraint(alignConstraint);
     },
 
     show: function() {
@@ -1028,7 +1054,6 @@ CalendarPopup.prototype = {
         // Reset the calendar to today's date
         this.calendar.setDate(new Date());
 
-        this._centerPopup();
         this.actor.lower(panelActor);
         this.actor.show();
         Tweener.addTween(this.actor,
@@ -1052,10 +1077,5 @@ CalendarPopup.prototype = {
                            onComplete: function() { this.actor.hide(); },
                            onCompleteScope: this
                          });
-    },
-
-    _centerPopup: function() {
-        let panelActor = Main.panel.actor;
-        this.actor.x = Math.round(panelActor.x + (panelActor.width - this.actor.width) / 2);
     }
 };
