@@ -1,4 +1,26 @@
 /* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */
+/*
+ * st-theme-node.c: style information for one node in a tree of themed objects
+ *
+ * Copyright 2008-2010 Red Hat, Inc.
+ * Copyright 2009 Steve Frécinaux
+ * Copyright 2009, 2010 Florian Müllner
+ * Copyright 2010 Adel Gadllah
+ * Copyright 2010 Giovanni Campagna
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU Lesser General Public License as
+ * published by the Free Software Foundation, either version 2.1 of
+ * the License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope it will be useful, but WITHOUT ANY
+ * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+ * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
+ * more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public License
+ * along with this program. If not, see <http://www.gnu.org/licenses/>.
+ */
 
 #include <stdlib.h>
 #include <string.h>
@@ -14,6 +36,9 @@ static void st_theme_node_finalize           (GObject                 *object);
 
 static const ClutterColor BLACK_COLOR = { 0, 0, 0, 0xff };
 static const ClutterColor TRANSPARENT_COLOR = { 0, 0, 0, 0 };
+static const ClutterColor DEFAULT_SUCCESS_COLOR = { 0x4e, 0x9a, 0x06, 0xff };
+static const ClutterColor DEFAULT_WARNING_COLOR = { 0xf5, 0x79, 0x3e, 0xff };
+static const ClutterColor DEFAULT_ERROR_COLOR = { 0xcc, 0x00, 0x00, 0xff };
 
 extern gfloat st_slow_down_factor;
 
@@ -99,10 +124,16 @@ st_theme_node_finalize (GObject *object)
 
   if (node->shadow)
     {
-      st_shadow_free (node->shadow);
+      st_shadow_unref (node->shadow);
       node->shadow = NULL;
     }
 
+  if (node->text_shadow)
+    {
+      st_shadow_unref (node->text_shadow);
+      node->text_shadow = NULL;
+    }
+
   if (node->background_image)
     g_free (node->background_image);
 
@@ -484,7 +515,7 @@ get_color_from_term (StThemeNode  *node,
 }
 
 /**
- * st_theme_node_get_color:
+ * st_theme_node_lookup_color:
  * @node: a #StThemeNode
  * @property_name: The name of the color property
  * @inherit: if %TRUE, if a value is not found for the property on the
@@ -501,14 +532,16 @@ get_color_from_term (StThemeNode  *node,
  * should be used instead. They are cached, so more efficient, and have
  * handling for shortcut properties and other details of CSS.
  *
+ * See also st_theme_node_get_color(), which provides a simpler API.
+ *
  * Return value: %TRUE if the property was found in the properties for this
  *  theme node (or in the properties of parent nodes when inheriting.)
  */
 gboolean
-st_theme_node_get_color (StThemeNode  *node,
-                         const char   *property_name,
-                         gboolean      inherit,
-                         ClutterColor *color)
+st_theme_node_lookup_color (StThemeNode  *node,
+                            const char   *property_name,
+                            gboolean      inherit,
+                            ClutterColor *color)
 {
 
   int i;
@@ -529,18 +562,51 @@ st_theme_node_get_color (StThemeNode  *node,
           else if (result == VALUE_INHERIT)
             {
               if (node->parent_node)
-                return st_theme_node_get_color (node->parent_node, property_name, inherit, color);
+                return st_theme_node_lookup_color (node->parent_node, property_name, inherit, color);
               else
                 break;
             }
         }
     }
 
+  if (inherit && node->parent_node)
+    return st_theme_node_lookup_color (node->parent_node, property_name, inherit, color);
+
   return FALSE;
 }
 
 /**
- * st_theme_node_get_double:
+ * st_theme_node_get_color:
+ * @node: a #StThemeNode
+ * @property_name: The name of the color property
+ * @color: location to store the color that was determined.
+ *
+ * Generically looks up a property containing a single color value. When
+ * specific getters (like st_theme_node_get_background_color()) exist, they
+ * should be used instead. They are cached, so more efficient, and have
+ * handling for shortcut properties and other details of CSS.
+ *
+ * If @property_name is not found, a warning will be logged and a
+ * default color returned.
+ *
+ * See also st_theme_node_lookup_color(), which provides more options,
+ * and lets you handle the case where the theme does not specify the
+ * indicated color.
+ */
+void
+st_theme_node_get_color (StThemeNode  *node,
+                         const char   *property_name,
+                         ClutterColor *color)
+{
+  if (!st_theme_node_lookup_color (node, property_name, FALSE, color))
+    {
+      g_warning ("Did not find color property '%s'", property_name);
+      memset (color, 0, sizeof (ClutterColor));
+    }
+}
+
+/**
+ * st_theme_node_lookup_double:
  * @node: a #StThemeNode
  * @property_name: The name of the numeric property
  * @inherit: if %TRUE, if a value is not found for the property on the
@@ -555,14 +621,16 @@ st_theme_node_get_color (StThemeNode  *node,
  * Generically looks up a property containing a single numeric value
  *  without units.
  *
+ * See also st_theme_node_get_double(), which provides a simpler API.
+ *
  * Return value: %TRUE if the property was found in the properties for this
  *  theme node (or in the properties of parent nodes when inheriting.)
  */
 gboolean
-st_theme_node_get_double (StThemeNode *node,
-                          const char  *property_name,
-                          gboolean     inherit,
-                          double      *value)
+st_theme_node_lookup_double (StThemeNode *node,
+                             const char  *property_name,
+                             gboolean     inherit,
+                             double      *value)
 {
   gboolean result = FALSE;
   int i;
@@ -587,11 +655,41 @@ st_theme_node_get_double (StThemeNode *node,
     }
 
   if (!result && inherit && node->parent_node)
-    result = st_theme_node_get_double (node->parent_node, property_name, inherit, value);
+    result = st_theme_node_lookup_double (node->parent_node, property_name, inherit, value);
 
   return result;
 }
 
+/**
+ * st_theme_node_get_double:
+ * @node: a #StThemeNode
+ * @property_name: The name of the numeric property
+ *
+ * Generically looks up a property containing a single numeric value
+ *  without units.
+ *
+ * See also st_theme_node_lookup_double(), which provides more options,
+ * and lets you handle the case where the theme does not specify the
+ * indicated value.
+ *
+ * Return value: the value found. If @property_name is not
+ *  found, a warning will be logged and 0 will be returned.
+ */
+gdouble
+st_theme_node_get_double (StThemeNode *node,
+                          const char  *property_name)
+{
+  gdouble value;
+
+  if (st_theme_node_lookup_double (node, property_name, FALSE, &value))
+    return value;
+  else
+    {
+      g_warning ("Did not find double property '%s'", property_name);
+      return 0.0;
+    }
+}
+
 static const PangoFontDescription *
 get_parent_font (StThemeNode *node)
 {
@@ -620,7 +718,7 @@ get_length_from_term (StThemeNode *node,
   if (term->type != TERM_NUMBER)
     {
       g_warning ("Ignoring length property that isn't a number");
-      return FALSE;
+      return VALUE_NOT_FOUND;
     }
 
   num = term->content.num;
@@ -679,8 +777,19 @@ get_length_from_term (StThemeNode *node,
       return VALUE_NOT_FOUND;
 
     case NUM_GENERIC:
-      g_warning ("length values must specify a unit");
-      return VALUE_NOT_FOUND;
+      {
+        if (num->val != 0)
+          {
+            g_warning ("length values must specify a unit");
+            return VALUE_NOT_FOUND;
+          }
+        else
+          {
+            type = ABSOLUTE;
+            multiplier = 0;
+          }
+        break;
+      }
 
     case NUM_PERCENTAGE:
       g_warning ("percentage lengths not currently supported");
@@ -750,7 +859,7 @@ get_length_from_term_int (StThemeNode *node,
   GetFromTermResult result;
 
   result = get_length_from_term (node, term, use_parent_font, &value);
-  if (result != VALUE_NOT_FOUND)
+  if (result == VALUE_FOUND)
     *length = (int) (0.5 + value);
   return result;
 }
@@ -782,7 +891,7 @@ get_length_internal (StThemeNode *node,
 }
 
 /**
- * st_theme_node_get_length:
+ * st_theme_node_lookup_length:
  * @node: a #StThemeNode
  * @property_name: The name of the length property
  * @inherit: if %TRUE, if a value is not found for the property on the
@@ -800,14 +909,16 @@ get_length_internal (StThemeNode *node,
  * should be used instead. They are cached, so more efficient, and have
  * handling for shortcut properties and other details of CSS.
  *
+ * See also st_theme_node_get_length(), which provides a simpler API.
+ *
  * Return value: %TRUE if the property was found in the properties for this
  *  theme node (or in the properties of parent nodes when inheriting.)
  */
 gboolean
-st_theme_node_get_length (StThemeNode *node,
-                          const char  *property_name,
-                          gboolean     inherit,
-                          gdouble     *length)
+st_theme_node_lookup_length (StThemeNode *node,
+                             const char  *property_name,
+                             gboolean     inherit,
+                             gdouble     *length)
 {
   GetFromTermResult result = get_length_internal (node, property_name, NULL, length);
   if (result == VALUE_FOUND)
@@ -816,12 +927,42 @@ st_theme_node_get_length (StThemeNode *node,
     inherit = TRUE;
 
   if (inherit && node->parent_node &&
-      st_theme_node_get_length (node->parent_node, property_name, inherit, length))
+      st_theme_node_lookup_length (node->parent_node, property_name, inherit, length))
     return TRUE;
   else
     return FALSE;
 }
 
+/**
+ * st_theme_node_get_length:
+ * @node: a #StThemeNode
+ * @property_name: The name of the length property
+ *
+ * Generically looks up a property containing a single length value. When
+ * specific getters (like st_theme_node_get_border_width()) exist, they
+ * should be used instead. They are cached, so more efficient, and have
+ * handling for shortcut properties and other details of CSS.
+ *
+ * Unlike st_theme_node_get_color() and st_theme_node_get_double(),
+ * this does not print a warning if the property is not found; it just
+ * returns 0.
+ *
+ * See also st_theme_node_lookup_length(), which provides more options.
+ *
+ * Return value: the length, in pixels, or 0 if the property was not found.
+ */
+gdouble
+st_theme_node_get_length (StThemeNode *node,
+                          const char  *property_name)
+{
+  gdouble length;
+
+  if (st_theme_node_lookup_length (node, property_name, FALSE, &length))
+    return length;
+  else
+    return 0.0;
+}
+
 static void
 do_border_radius_term (StThemeNode *node,
                        CRTerm      *term,
@@ -912,7 +1053,7 @@ do_border_property (StThemeNode   *node,
   StSide side = (StSide)-1;
   ClutterColor color;
   gboolean color_set = FALSE;
-  int width;
+  int width = 0; /* suppress warning */
   gboolean width_set = FALSE;
   int j;
 
@@ -957,7 +1098,8 @@ do_border_property (StThemeNode   *node,
               const char *ident = term->content.str->stryng->str;
               if (strcmp (ident, "none") == 0 || strcmp (ident, "hidden") == 0)
                 {
-                  width = 0.;
+                  width = 0;
+                  width_set = TRUE;
                   continue;
                 }
               else if (strcmp (ident, "solid") == 0)
@@ -1044,7 +1186,7 @@ do_outline_property (StThemeNode   *node,
   const char *property_name = decl->property->stryng->str + 7; /* Skip 'outline' */
   ClutterColor color;
   gboolean color_set = FALSE;
-  int width;
+  int width = 0; /* suppress warning */
   gboolean width_set = FALSE;
 
   if (strcmp (property_name, "") == 0)
@@ -1061,7 +1203,8 @@ do_outline_property (StThemeNode   *node,
               const char *ident = term->content.str->stryng->str;
               if (strcmp (ident, "none") == 0 || strcmp (ident, "hidden") == 0)
                 {
-                  width = 0.;
+                  width = 0;
+                  width_set = TRUE;
                   continue;
                 }
               else if (strcmp (ident, "solid") == 0)
@@ -1424,6 +1567,7 @@ _st_theme_node_ensure_background (StThemeNode *node)
   node->background_computed = TRUE;
   node->background_color = TRANSPARENT_COLOR;
   node->background_gradient_type = ST_GRADIENT_NONE;
+  node->background_position_set = FALSE;
 
   ensure_properties (node);
 
@@ -1450,6 +1594,7 @@ _st_theme_node_ensure_background (StThemeNode *node)
           node->background_color = TRANSPARENT_COLOR;
           g_free (node->background_image);
           node->background_image = NULL;
+          node->background_position_set = FALSE;
 
           for (term = decl->value; term; term = term->next)
             {
@@ -1472,12 +1617,40 @@ _st_theme_node_ensure_background (StThemeNode *node)
                 }
               else if (term->type == TERM_URI)
                 {
+                  CRStyleSheet *base_stylesheet;
+
+                  if (decl->parent_statement != NULL)
+                    base_stylesheet = decl->parent_statement->parent_sheet;
+                  else
+                    base_stylesheet = NULL;
+
                   node->background_image = _st_theme_resolve_url (node->theme,
-                                                                     decl->parent_statement->parent_sheet,
-                                                                     term->content.str->stryng->str);
+                                                                  base_stylesheet,
+                                                                  term->content.str->stryng->str);
                 }
             }
         }
+      else if (strcmp (property_name, "-position") == 0)
+        {
+          GetFromTermResult result = get_length_from_term_int (node, decl->value, FALSE, &node->background_position_x);
+          if (result == VALUE_NOT_FOUND)
+            {
+              node->background_position_set = FALSE;
+              continue;
+            }
+          else
+            node->background_position_set = TRUE;
+            
+           result = get_length_from_term_int (node, decl->value->next, FALSE, &node->background_position_y);
+
+          if (result == VALUE_NOT_FOUND)
+            {
+              node->background_position_set = FALSE;
+              continue;
+            }
+          else
+            node->background_position_set = TRUE;
+        }
       else if (strcmp (property_name, "-color") == 0)
         {
           GetFromTermResult result;
@@ -1503,9 +1676,16 @@ _st_theme_node_ensure_background (StThemeNode *node)
 
           if (decl->value->type == TERM_URI)
             {
+              CRStyleSheet *base_stylesheet;
+
+              if (decl->parent_statement != NULL)
+                base_stylesheet = decl->parent_statement->parent_sheet;
+              else
+                base_stylesheet = NULL;
+
               g_free (node->background_image);
               node->background_image = _st_theme_resolve_url (node->theme,
-                                                              decl->parent_statement->parent_sheet,
+                                                              base_stylesheet,
                                                               decl->value->content.str->stryng->str);
             }
           else if (term_is_inherit (decl->value))
@@ -1687,7 +1867,7 @@ st_theme_node_get_transition_duration (StThemeNode *node)
   if (node->transition_duration > -1)
     return st_slow_down_factor * node->transition_duration;
 
-  st_theme_node_get_double (node, "transition-duration", FALSE, &value);
+  st_theme_node_lookup_double (node, "transition-duration", FALSE, &value);
 
   node->transition_duration = (int)value;
 
@@ -2070,17 +2250,19 @@ font_variant_from_term (CRTerm       *term,
 const PangoFontDescription *
 st_theme_node_get_font (StThemeNode *node)
 {
-  PangoStyle font_style;
+  /* Initialized despite _set flags to suppress compiler warnings */
+  PangoStyle font_style = PANGO_STYLE_NORMAL;
   gboolean font_style_set = FALSE;
-  PangoVariant variant;
+  PangoVariant variant = PANGO_VARIANT_NORMAL;
   gboolean variant_set = FALSE;
-  PangoWeight weight;
-  gboolean weight_absolute;
+  PangoWeight weight = PANGO_WEIGHT_NORMAL;
+  gboolean weight_absolute = TRUE;
   gboolean weight_set = FALSE;
-  double parent_size;
-  double size = 0.; /* Suppress warning */
+  double size = 0.;
   gboolean size_set = FALSE;
+
   char *family = NULL;
+  double parent_size;
   int i;
 
   if (node->font_desc)
@@ -2287,6 +2469,7 @@ st_theme_node_get_border_image (StThemeNode *node)
       if (strcmp (decl->property->stryng->str, "border-image") == 0)
         {
           CRTerm *term = decl->value;
+          CRStyleSheet *base_stylesheet;
           int borders[4];
           int n_borders = 0;
           int i;
@@ -2361,7 +2544,12 @@ st_theme_node_get_border_image (StThemeNode *node)
               break;
             }
 
-          filename = _st_theme_resolve_url (node->theme, decl->parent_statement->parent_sheet, url);
+          if (decl->parent_statement != NULL)
+            base_stylesheet = decl->parent_statement->parent_sheet;
+          else
+            base_stylesheet = NULL;
+
+          filename = _st_theme_resolve_url (node->theme, base_stylesheet, url);
           if (filename == NULL)
             goto next_property;
 
@@ -2418,6 +2606,72 @@ st_theme_node_get_vertical_padding (StThemeNode *node)
   return padding;
 }
 
+static void
+parse_shadow_property (StThemeNode   *node,
+                       CRDeclaration *decl,
+                       ClutterColor  *color,
+                       gdouble       *xoffset,
+                       gdouble       *yoffset,
+                       gdouble       *blur,
+                       gdouble       *spread)
+{
+  /* Set value for width/color/blur in any order */
+  CRTerm *term;
+  int n_offsets = 0;
+
+  /* default values */
+  color->red = 0x0; color->green = 0x0; color->blue = 0x0; color->alpha = 0xff;
+  *xoffset = 0.;
+  *yoffset = 0.;
+  *blur = 0.;
+  *spread = 0.;
+
+  for (term = decl->value; term; term = term->next)
+    {
+      GetFromTermResult result;
+
+      if (term->type == TERM_NUMBER)
+        {
+          gdouble value;
+          gdouble multiplier;
+
+          multiplier = (term->unary_op == MINUS_UOP) ? -1. : 1.;
+          result = get_length_from_term (node, term, FALSE, &value);
+          if (result != VALUE_NOT_FOUND)
+            {
+              switch (n_offsets++)
+                {
+                case 0:
+                  *xoffset = multiplier * value;
+                  break;
+                case 1:
+                  *yoffset = multiplier * value;
+                  break;
+                case 2:
+                  if (multiplier < 0)
+                      g_warning ("Negative blur values are "
+                                 "not allowed");
+                  *blur = value;
+                  break;
+                case 3:
+                  if (multiplier < 0)
+                      g_warning ("Negative spread values are "
+                                 "not allowed");
+                  *spread = value;
+                  break;
+                }
+              continue;
+            }
+        }
+
+      result = get_color_from_term (node, term, color);
+      if (result != VALUE_NOT_FOUND)
+        {
+          continue;
+        }
+    }
+}
+
 /**
  * st_theme_node_get_shadow:
  * @node: a #StThemeNode
@@ -2446,59 +2700,15 @@ st_theme_node_get_shadow (StThemeNode *node)
 
       if (strcmp (decl->property->stryng->str, "-st-shadow") == 0)
         {
-          /* Set value for width/color/blur in any order */
-          CRTerm *term;
-          ClutterColor color = { 0x0, 0x0, 0x0, 0xff };
-          gdouble xoffset = 0.;
-          gdouble yoffset = 0.;
-          gdouble blur = 0.;
-          gdouble spread = 0.;
-          int n_offsets = 0;
+          ClutterColor color;
+          gdouble xoffset;
+          gdouble yoffset;
+          gdouble blur;
+          gdouble spread;
 
-          for (term = decl->value; term; term = term->next)
-            {
-              GetFromTermResult result;
+          parse_shadow_property (node, decl,
+                                 &color, &xoffset, &yoffset, &blur, &spread);
 
-              if (term->type == TERM_NUMBER)
-                {
-                  gdouble value;
-                  gdouble multiplier;
-
-                  multiplier = (term->unary_op == MINUS_UOP) ? -1. : 1.;
-                  result = get_length_from_term (node, term, FALSE, &value);
-                  if (result != VALUE_NOT_FOUND)
-                    {
-                      switch (n_offsets++)
-                        {
-                        case 0:
-                          xoffset = multiplier * value;
-                          break;
-                        case 1:
-                          yoffset = multiplier * value;
-                          break;
-                        case 2:
-                          if (multiplier < 0)
-                              g_warning ("Negative blur values are "
-                                         "not allowed");
-                          blur = value;
-                          break;
-                        case 3:
-                          if (multiplier < 0)
-                              g_warning ("Negative spread values are "
-                                         "not allowed");
-                          spread = value;
-                          break;
-                        }
-                      continue;
-                    }
-                }
-
-              result = get_color_from_term (node, term, &color);
-              if (result != VALUE_NOT_FOUND)
-                {
-                  continue;
-                }
-            }
           node->shadow = st_shadow_new (&color,
                                         xoffset, yoffset,
                                         blur, spread);
@@ -2509,6 +2719,181 @@ st_theme_node_get_shadow (StThemeNode *node)
   return NULL;
 }
 
+/**
+ * st_theme_node_get_text_shadow:
+ * @node: a #StThemeNode
+ *
+ * Gets the value for the text-shadow style property
+ *
+ * Return value: (transfer none): the node's text-shadow, or %NULL
+ *   if node has no text-shadow
+ */
+StShadow *
+st_theme_node_get_text_shadow (StThemeNode *node)
+{
+  StShadow *result = NULL;
+  int i;
+
+  if (node->text_shadow_computed)
+    return node->text_shadow;
+
+  ensure_properties (node);
+
+  for (i = node->n_properties - 1; i >= 0; i--)
+    {
+      CRDeclaration *decl = node->properties[i];
+
+      if (strcmp (decl->property->stryng->str, "text-shadow") == 0)
+        {
+          ClutterColor color;
+          gdouble xoffset;
+          gdouble yoffset;
+          gdouble blur;
+          gdouble spread;
+
+          parse_shadow_property (node, decl,
+                                 &color, &xoffset, &yoffset, &blur, &spread);
+
+          result = st_shadow_new (&color,
+                                  xoffset, yoffset,
+                                  blur, spread);
+
+          break;
+        }
+    }
+
+  if (!result && node->parent_node)
+    {
+      result = st_theme_node_get_text_shadow (node->parent_node);
+      if (result)
+        st_shadow_ref (result);
+    }
+
+  node->text_shadow = result;
+  node->text_shadow_computed = TRUE;
+
+  return result;
+}
+
+/**
+ * st_theme_node_get_icon_colors:
+ * @node: a #StThemeNode
+ *
+ * Gets the colors that should be used for colorizing symbolic icons according
+ * the style of this node.
+ *
+ * Return value: (transfer none): the icon colors to use for this theme node
+ */
+StIconColors *
+st_theme_node_get_icon_colors (StThemeNode *node)
+{
+  /* Foreground here will always be the same as st_theme_node_get_foreground_color(),
+   * but there's a loss of symmetry and little efficiency win if we try to exploit
+   * that. */
+
+  enum {
+    FOREGROUND = 1 << 0,
+    WARNING = 1 << 1,
+    ERROR = 1 << 2,
+    SUCCESS = 1 << 3
+  };
+
+  gboolean shared_with_parent;
+  int i;
+  ClutterColor color = { 0, };
+
+  guint still_need = FOREGROUND | WARNING | ERROR | SUCCESS;
+
+  g_return_val_if_fail (ST_IS_THEME_NODE (node), NULL);
+
+  if (node->icon_colors)
+    return node->icon_colors;
+
+  if (node->parent_node)
+    {
+      node->icon_colors = st_theme_node_get_icon_colors (node->parent_node);
+      shared_with_parent = TRUE;
+    }
+  else
+    {
+      node->icon_colors = st_icon_colors_new ();
+      node->icon_colors->foreground = BLACK_COLOR;
+      node->icon_colors->warning = DEFAULT_WARNING_COLOR;
+      node->icon_colors->error = DEFAULT_ERROR_COLOR;
+      node->icon_colors->success = DEFAULT_SUCCESS_COLOR;
+      shared_with_parent = FALSE;
+    }
+
+  ensure_properties (node);
+
+  for (i = node->n_properties - 1; i >= 0 && still_need != 0; i--)
+    {
+      CRDeclaration *decl = node->properties[i];
+      GetFromTermResult result = VALUE_NOT_FOUND;
+      guint found = 0;
+
+      if ((still_need & FOREGROUND) != 0 &&
+          strcmp (decl->property->stryng->str, "color") == 0)
+        {
+          found = FOREGROUND;
+          result = get_color_from_term (node, decl->value, &color);
+        }
+      else if ((still_need & WARNING) != 0 &&
+               strcmp (decl->property->stryng->str, "warning-color") == 0)
+        {
+          found = WARNING;
+          result = get_color_from_term (node, decl->value, &color);
+        }
+      else if ((still_need & ERROR) != 0 &&
+               strcmp (decl->property->stryng->str, "error-color") == 0)
+        {
+          found = ERROR;
+          result = get_color_from_term (node, decl->value, &color);
+        }
+      else if ((still_need & SUCCESS) != 0 &&
+               strcmp (decl->property->stryng->str, "success-color") == 0)
+        {
+          found = SUCCESS;
+          result = get_color_from_term (node, decl->value, &color);
+        }
+
+      if (result == VALUE_INHERIT)
+        {
+          still_need &= ~found;
+        }
+      else if (result == VALUE_FOUND)
+        {
+          still_need &= ~found;
+          if (shared_with_parent)
+            {
+              node->icon_colors = st_icon_colors_copy (node->icon_colors);
+              shared_with_parent = FALSE;
+            }
+
+          switch (found)
+            {
+            case FOREGROUND:
+              node->icon_colors->foreground = color;
+              break;
+            case WARNING:
+              node->icon_colors->warning = color;
+              break;
+            case ERROR:
+              node->icon_colors->error = color;
+              break;
+            case SUCCESS:
+              node->icon_colors->success = color;
+              break;
+            }
+        }
+    }
+
+  if (shared_with_parent)
+    st_icon_colors_ref (node->icon_colors);
+
+  return node->icon_colors;
+}
+
 static float
 get_width_inc (StThemeNode *node)
 {
@@ -2551,8 +2936,8 @@ st_theme_node_adjust_for_height (StThemeNode  *node,
 /**
  * st_theme_node_adjust_preferred_width:
  * @node: a #StThemeNode
- * @min_width_p: (inout) (allow-none): the width to adjust
- * @for_height: (inout): the height to adjust
+ * @min_width_p: (inout) (allow-none): the minimum width to adjust
+ * @natural_width_p: (inout): the natural width to adjust
  *
  * Adjusts the minimum and natural width computed for an actor by
  * adding on the necessary space for borders and padding and taking
@@ -2598,7 +2983,7 @@ st_theme_node_adjust_preferred_width (StThemeNode  *node,
  * Adjusts a "for width" passed to clutter_actor_get_preferred_height() to
  * account for borders and padding. This is a convenience function meant
  * to be called from a get_preferred_height() method of a #ClutterActor
- * subclass. The value after adjustmnet is the width available for the actor's
+ * subclass. The value after adjustment is the width available for the actor's
  * content.
  */
 void
@@ -2618,8 +3003,8 @@ st_theme_node_adjust_for_width (StThemeNode  *node,
 /**
  * st_theme_node_adjust_preferred_height:
  * @node: a #StThemeNode
- * @min_height_p: (inout) (allow-none): the height to adjust
- * @for_height: (inout): the height to adjust
+ * @min_height_p: (inout) (allow-none): the minimum height to adjust
+ * @natural_height_p: (inout): the natural height to adjust
  *
  * Adjusts the minimum and natural height computed for an actor by
  * adding on the necessary space for borders and padding and taking
@@ -2760,6 +3145,9 @@ st_theme_node_geometry_equal (StThemeNode *node,
 {
   StSide side;
 
+  g_return_val_if_fail (ST_IS_THEME_NODE (node), FALSE);
+  g_return_val_if_fail (ST_IS_THEME_NODE (other), FALSE);
+
   _st_theme_node_ensure_geometry (node);
   _st_theme_node_ensure_geometry (other);
 
@@ -2780,3 +3168,86 @@ st_theme_node_geometry_equal (StThemeNode *node,
 
   return TRUE;
 }
+
+/**
+ * st_theme_node_paint_equal:
+ * @node: a #StThemeNode
+ * @other: a different #StThemeNode
+ *
+ * Check if st_theme_node_paint() will paint identically for @node as it does
+ * for @other. Note that in some cases this function may return %TRUE even
+ * if there is no visible difference in the painting.
+ *
+ * Return value: %TRUE if the two theme nodes paint identically. %FALSE if the
+ *   two nodes potentially paint differently.
+ */
+gboolean
+st_theme_node_paint_equal (StThemeNode *node,
+                           StThemeNode *other)
+{
+  StBorderImage *border_image, *other_border_image;
+  StShadow *shadow, *other_shadow;
+  int i;
+
+  g_return_val_if_fail (ST_IS_THEME_NODE (node), FALSE);
+  g_return_val_if_fail (ST_IS_THEME_NODE (other), FALSE);
+
+  _st_theme_node_ensure_background (node);
+  _st_theme_node_ensure_background (other);
+
+  if (!clutter_color_equal (&node->background_color, &other->background_color))
+    return FALSE;
+
+  if (node->background_gradient_type != other->background_gradient_type)
+    return FALSE;
+
+  if (node->background_gradient_type != ST_GRADIENT_NONE &&
+      !clutter_color_equal (&node->background_gradient_end, &other->background_gradient_end))
+    return FALSE;
+
+  if (g_strcmp0 (node->background_image, other->background_image) != 0)
+    return FALSE;
+
+  _st_theme_node_ensure_geometry (node);
+  _st_theme_node_ensure_geometry (other);
+
+  for (i = 0; i < 4; i++)
+    {
+      if (node->border_width[i] != other->border_width[i])
+        return FALSE;
+
+      if (node->border_width[i] > 0 &&
+          !clutter_color_equal (&node->border_color[i], &other->border_color[i]))
+        return FALSE;
+
+      if (node->border_radius[i] != other->border_radius[i])
+        return FALSE;
+    }
+
+  if (node->outline_width != other->outline_width)
+    return FALSE;
+
+  if (node->outline_width > 0 &&
+      !clutter_color_equal (&node->outline_color, &other->outline_color))
+    return FALSE;
+
+  border_image = st_theme_node_get_border_image (node);
+  other_border_image = st_theme_node_get_border_image (other);
+
+  if ((border_image == NULL) != (other_border_image == NULL))
+    return FALSE;
+
+  if (border_image != NULL && !st_border_image_equal (border_image, other_border_image))
+    return FALSE;
+
+  shadow = st_theme_node_get_shadow (node);
+  other_shadow = st_theme_node_get_shadow (other);
+
+  if ((shadow == NULL) != (other_shadow == NULL))
+    return FALSE;
+
+  if (shadow != NULL && !st_shadow_equal (shadow, other_shadow))
+    return FALSE;
+
+  return TRUE;
+}
